EXPORTS
	kdc_log
	kdc_log_msg
	kdc_log_msg_va
	kdc_openlog
	krb5_kdc_windc_init
	krb5_kdc_get_config
	krb5_kdc_pkinit_config
	krb5_kdc_set_dbinfo
	krb5_kdc_process_krb5_request
	krb5_kdc_process_request
	krb5_kdc_save_request
	krb5_kdc_update_time
	krb5_kdc_pk_initialize
