EXPORTS
	__gss_c_nt_anonymous_oid_desc	DATA
	__gss_c_nt_export_name_oid_desc	DATA
	__gss_c_nt_hostbased_service_oid_desc	DATA
	__gss_c_nt_hostbased_service_x_oid_desc	DATA
	__gss_c_nt_machine_uid_name_oid_desc	DATA
	__gss_c_nt_string_uid_name_oid_desc	DATA
	__gss_c_nt_user_name_oid_desc	DATA
	__gss_krb5_nt_principal_name_oid_desc	DATA
	__gss_c_attr_stream_sizes_oid_desc	DATA
        __gss_c_attr_local_login_user           DATA
        __gss_c_cred_certificate_oid_desc       DATA
        __gss_c_cred_password_oid_desc          DATA
	gss_accept_sec_context
	gss_acquire_cred
	gss_acquire_cred_with_password
	gss_add_buffer_set_member
	gss_add_cred
	gss_add_cred_with_password
	gss_add_oid_set_member
        gss_authorize_localname
	gss_canonicalize_name
	gss_compare_name
	gss_context_query_attributes
	gss_context_time
	gss_create_empty_buffer_set
	gss_create_empty_oid_set
	gss_decapsulate_token
	gss_delete_name_attribute
	gss_delete_sec_context
        gss_display_mech_attr
	gss_display_name
	gss_display_name_ext
	gss_display_status
	gss_duplicate_name
	gss_duplicate_oid
	gss_encapsulate_token
        gss_export_cred
	gss_export_name
	gss_export_name_composite
	gss_export_sec_context
	gss_get_mic
	gss_get_name_attribute
        gss_import_cred
	gss_import_name
	gss_import_sec_context
	gss_indicate_mechs
        gss_indicate_mechs_by_attrs
	gss_init_sec_context
        gss_inquire_attrs_for_mech
	gss_inquire_context
	gss_inquire_cred
	gss_inquire_cred_by_mech
	gss_inquire_cred_by_oid
        gss_inquire_mech_for_saslname
	gss_inquire_mechs_for_name
	gss_inquire_name
	gss_inquire_names_for_mech
        gss_inquire_saslname_for_mech
	gss_inquire_sec_context_by_oid  ;!
	gss_krb5_ccache_name
	gss_krb5_copy_ccache
	gss_krb5_export_lucid_sec_context
	gss_krb5_free_lucid_sec_context
	gss_krb5_get_tkt_flags
	gss_krb5_import_cred
	gss_krb5_set_allowable_enctypes
	gss_localname
	gss_mg_collect_error
        gss_mo_get
        gss_mo_set
        gss_mo_list
        gss_mo_name
        gss_name_to_oid
        gss_oid_to_name
	gss_oid_equal
	gss_oid_to_str
	gss_pname_to_uid
	gss_process_context_token
	gss_pseudo_random
	gss_release_buffer
	gss_release_buffer_set
	gss_release_cred
	gss_release_iov_buffer
	gss_release_name
	gss_release_oid
	gss_release_oid_set
	gss_seal
	gss_set_cred_option
	gss_set_name_attribute
	gss_set_sec_context_option
	gss_sign
	gss_store_cred
	gss_test_oid_set_member
	gss_unseal
	gss_unwrap
	gss_unwrap_aead
	gss_unwrap_iov
	gss_userok
	gss_verify
	gss_verify_mic
	gss_wrap
	gss_wrap_aead
	gss_wrap_iov
	gss_wrap_iov_length
	gss_wrap_size_limit
	gsskrb5_extract_authtime_from_sec_context
	gsskrb5_extract_authz_data_from_sec_context
	gsskrb5_extract_service_keyblock
	gsskrb5_get_initiator_subkey
	gsskrb5_get_subkey
	gsskrb5_get_time_offset
	gsskrb5_register_acceptor_identity
	gsskrb5_set_default_realm
	gsskrb5_set_dns_canonicalize
	gsskrb5_set_send_to_kdc
	gsskrb5_set_time_offset
	krb5_gss_register_acceptor_identity

; _gsskrb5cfx_ are really internal symbols, but export
; then now to make testing easier.
	_gsskrb5cfx_wrap_length_cfx
	_gssapi_wrap_size_cfx

        initialize_gk5_error_table_r    ;!

	__gss_krb5_copy_ccache_x_oid_desc	DATA
	__gss_krb5_get_tkt_flags_x_oid_desc	DATA
	__gss_krb5_extract_authz_data_from_sec_context_x_oid_desc	DATA
	__gss_krb5_compat_des3_mic_x_oid_desc	DATA
	__gss_krb5_register_acceptor_identity_x_oid_desc	DATA
	__gss_krb5_export_lucid_context_x_oid_desc	DATA
	__gss_krb5_export_lucid_context_v1_x_oid_desc	DATA
	__gss_krb5_set_dns_canonicalize_x_oid_desc	DATA
	__gss_krb5_get_subkey_x_oid_desc	DATA
	__gss_krb5_get_initiator_subkey_x_oid_desc	DATA
	__gss_krb5_get_acceptor_subkey_x_oid_desc	DATA
	__gss_krb5_send_to_kdc_x_oid_desc	DATA
	__gss_krb5_get_authtime_x_oid_desc	DATA
	__gss_krb5_get_service_keyblock_x_oid_desc	DATA
	__gss_krb5_set_allowable_enctypes_x_oid_desc	DATA
	__gss_krb5_set_default_realm_x_oid_desc	DATA
	__gss_krb5_ccache_name_x_oid_desc	DATA
	__gss_krb5_set_time_offset_x_oid_desc	DATA
	__gss_krb5_get_time_offset_x_oid_desc	DATA
	__gss_krb5_plugin_register_x_oid_desc	DATA
	__gss_ntlm_get_session_key_x_oid_desc	DATA
	__gss_c_nt_ntlm_oid_desc	DATA
	__gss_c_nt_dn_oid_desc	DATA
	__gss_krb5_nt_principal_name_referral_oid_desc	DATA
	__gss_c_ntlm_avguest_oid_desc	DATA
	__gss_c_ntlm_v1_oid_desc	DATA
	__gss_c_ntlm_v2_oid_desc	DATA
	__gss_c_ntlm_session_key_oid_desc	DATA
	__gss_c_ntlm_force_v1_oid_desc	DATA
	__gss_krb5_cred_no_ci_flags_x_oid_desc	DATA
	__gss_krb5_import_cred_x_oid_desc	DATA
	__gss_c_ma_sasl_mech_name_oid_desc	DATA
	__gss_c_ma_mech_name_oid_desc	DATA
	__gss_c_ma_mech_description_oid_desc	DATA
	__gss_sasl_digest_md5_mechanism_oid_desc	DATA
	__gss_krb5_mechanism_oid_desc	DATA
	__gss_ntlm_mechanism_oid_desc	DATA
	__gss_spnego_mechanism_oid_desc	DATA
	__gss_c_peer_has_updated_spnego_oid_desc	DATA
	__gss_c_ma_mech_concrete_oid_desc	DATA
	__gss_c_ma_mech_pseudo_oid_desc	DATA
	__gss_c_ma_mech_composite_oid_desc	DATA
	__gss_c_ma_mech_nego_oid_desc	DATA
	__gss_c_ma_mech_glue_oid_desc	DATA
	__gss_c_ma_not_mech_oid_desc	DATA
	__gss_c_ma_deprecated_oid_desc	DATA
	__gss_c_ma_not_dflt_mech_oid_desc	DATA
	__gss_c_ma_itok_framed_oid_desc	DATA
	__gss_c_ma_auth_init_oid_desc	DATA
	__gss_c_ma_auth_targ_oid_desc	DATA
	__gss_c_ma_auth_init_init_oid_desc	DATA
	__gss_c_ma_auth_targ_init_oid_desc	DATA
	__gss_c_ma_auth_init_anon_oid_desc	DATA
	__gss_c_ma_auth_targ_anon_oid_desc	DATA
	__gss_c_ma_deleg_cred_oid_desc	DATA
	__gss_c_ma_integ_prot_oid_desc	DATA
	__gss_c_ma_conf_prot_oid_desc	DATA
	__gss_c_ma_mic_oid_desc	DATA
	__gss_c_ma_wrap_oid_desc	DATA
	__gss_c_ma_prot_ready_oid_desc	DATA
	__gss_c_ma_replay_det_oid_desc	DATA
	__gss_c_ma_oos_det_oid_desc	DATA
	__gss_c_ma_cbindings_oid_desc	DATA
	__gss_c_ma_pfs_oid_desc	DATA
	__gss_c_ma_compress_oid_desc	DATA
	__gss_c_ma_ctx_trans_oid_desc	DATA
